import dataclasses
import datetime
import itertools
import re
import time
from pathlib import Path
from threading import Lock
from types import MappingProxyType
from typing import Any, Iterable, Mapping

import diskcache
import funcy
import loguru
import platformdirs
import reasonable
import yaml_ld
from nanopub import NanopubClient
from rdflib import ConjunctiveGraph, Namespace, URIRef, Variable
from rdflib.plugins.sparql.algebra import translateQuery
from rdflib.plugins.sparql.evaluate import evalQuery
from rdflib.plugins.sparql.parser import parseQuery
from rdflib.plugins.sparql.parserutils import CompValue
from rdflib.plugins.sparql.sparql import Query
from rdflib.query import Processor
from rdflib.term import BNode, Literal, Node
from requests.exceptions import ConnectionError
from yaml_ld.document_loaders.content_types import ParserNotFound
from yaml_ld.errors import NotFound, YAMLLDError
from yarl import URL

from iolanta.errors import UnresolvedIRI
from iolanta.namespaces import (  # noqa: WPS235
    DC,
    DCTERMS,
    FOAF,
    IOLANTA,
    OWL,
    PROV,
    RDF,
    RDFS,
    VANN,
)
from iolanta.parse_quads import parse_quads

NORMALIZE_TERMS_MAP = MappingProxyType({
    URIRef(_url := 'https://www.w3.org/2002/07/owl'): URIRef(f'{_url}#'),
    URIRef(_url := 'https://www.w3.org/2000/01/rdf-schema'): URIRef(f'{_url}#'),
})


REASONING_ENABLED = True
OWL_REASONING_ENABLED = False
INDICES = [
    URIRef('https://iolanta.tech/visualizations/index.yaml'),
]


REDIRECTS = MappingProxyType({
    # FIXME This is presently hardcoded; we need to
    #   - either find a way to resolve these URLs automatically,
    #   - or create a repository of those redirects online.
    'https://purl.org/vocab/vann/': URIRef(
        'https://vocab.org/vann/vann-vocab-20100607.rdf',
    ),
    URIRef(DC): URIRef(DCTERMS),
    URIRef(RDF): URIRef(RDF),
    URIRef(RDFS): URIRef(RDFS),
    URIRef(OWL): URIRef(OWL),

    # This one does not answer via HTTPS :(
    URIRef('https://xmlns.com/foaf/0.1/'): URIRef('http://xmlns.com/foaf/0.1/'),
    URIRef('https://www.nanopub.org/nschema'): URIRef(
        'https://www.nanopub.net/nschema#',
    ),
    URIRef('https://nanopub.org/nschema'): URIRef(
        'https://nanopub.net/nschema#',
    ),
    URIRef(PROV): URIRef('https://www.w3.org/ns/prov-o'),
})


@diskcache.Cache(
    directory=str(
        platformdirs.user_cache_path(
            appname='iolanta',
        ) / 'find_retractions_for',
    ),
).memoize(expire=datetime.timedelta(days=8).total_seconds())
def find_retractions_for(nanopublication: URIRef) -> set[URIRef]:
    """Find nanopublications that retract the given one."""
    # See https://github.com/fair-workflows/nanopub/issues/168 for
    # context of this dirty hack.
    use_server = 'http://grlc.nanopubs.lod.labs.vu.nl/api/local/local/'

    client = NanopubClient(
        use_server=use_server,
    )
    client.grlc_urls = [use_server]

    retractions = client.find_retractions_of(
        str(nanopublication).replace(
            'https://',
            'http://',
        ),
    )

    return {URIRef(retraction) for retraction in retractions}


def _extract_from_mapping(  # noqa: WPS213
    algebra: Mapping[str, Any],
) -> Iterable[URIRef | Variable]:
    match algebra.name:
        case 'SelectQuery' | 'Project' | 'Distinct':
            yield from extract_mentioned_urls(algebra['p'])

        case 'BGP':
            yield from [   # noqa: WPS353, WPS221
                term
                for triple in algebra['triples']
                for term in triple
                if not isinstance(term, Literal)
            ]

        case 'Filter' | 'UnaryNot' | 'OrderCondition':
            yield from extract_mentioned_urls(algebra['expr'])   # noqa: WPS204

        case built_in if built_in.startswith('Builtin_'):
            yield from extract_mentioned_urls(algebra['arg'])

        case 'RelationalExpression':
            yield from extract_mentioned_urls(algebra['expr'])
            yield from extract_mentioned_urls(algebra['other'])

        case 'LeftJoin':
            yield from extract_mentioned_urls(algebra['p1'])
            yield from extract_mentioned_urls(algebra['p2'])
            yield from extract_mentioned_urls(algebra['expr'])

        case 'Join':
            yield from extract_mentioned_urls(algebra['p1'])
            yield from extract_mentioned_urls(algebra['p2'])

        case 'ConditionalOrExpression' | 'ConditionalAndExpression':
            yield from extract_mentioned_urls(algebra['expr'])
            yield from extract_mentioned_urls(algebra['other'])

        case 'OrderBy':
            yield from extract_mentioned_urls(algebra['p'])
            yield from extract_mentioned_urls(algebra['expr'])

        case 'TrueFilter':
            return

        case 'Graph':
            yield from extract_mentioned_urls(algebra['p'])
            yield from extract_mentioned_urls(algebra['term'])

        case unknown_name:
            formatted_keys = ', '.join(algebra.keys())
            loguru.logger.error(
                'Unknown SPARQL expression '
                f'{unknown_name}({formatted_keys}): {algebra}',
            )
            return


def extract_mentioned_urls(
    algebra,
) -> Iterable[URIRef | Variable]:
    """Extract flat triples from parsed SPARQL query."""
    match algebra:
        case Variable() as query_variable:
            yield query_variable

        case URIRef() as uri_ref:
            yield uri_ref

        case dict():
            yield from _extract_from_mapping(algebra)

        case list() as expressions:
            for expression in expressions:
                yield from extract_mentioned_urls(expression)

        case unknown_algebra:
            algebra_type = type(unknown_algebra)
            raise ValueError(
                f'Algebra of unknown type {algebra_type}: {unknown_algebra}',
            )


def normalize_term(term: Node) -> Node:
    """
    Normalize RDF terms.

    This is an exctremely dirty hack to fix a bug in OWL reported here:

    > https://stackoverflow.com/q/78934864/1245471

    TODO This is:
      * A dirty hack;
      * Based on hard code.
    """
    if isinstance(term, URIRef) and term.startswith('http://'):
        term = URIRef(re.sub('^http', 'https', term))

    return NORMALIZE_TERMS_MAP.get(term, term)


def resolve_variables(
    terms: Iterable[URIRef | Variable],
    bindings: Mapping[str, Node],
):
    """Replace variables with their values."""
    for term in terms:
        match term:
            case URIRef():
                yield term

            case Variable() as query_variable:
                variable_value = bindings.get(str(query_variable))
                if (
                    variable_value is not None
                    and isinstance(variable_value, URIRef)
                ):
                    yield variable_value


def extract_mentioned_urls_from_query(
    query: Query,
    bindings: dict[str, Node],
    base: str | None,
    namespaces: dict[str, Namespace],
) -> tuple[Query, set[URIRef]]:
    """Extract URLs that a SPARQL query somehow mentions."""
    return query, set(
        resolve_variables(
            extract_mentioned_urls(query.algebra),
            bindings=bindings,
        ),
    )


@dataclasses.dataclass
class Loaded:
    """The data was loaded successfully."""


@dataclasses.dataclass
class Skipped:
    """The data is already in the graph and loading was skipped."""


LoadResult = Loaded | Skipped


def _extract_nanopublication_uris(
    algebra: CompValue,
) -> Iterable[URIRef]:
    """Extract nanopublications to get retracting information for."""
    match algebra.name:
        case 'SelectQuery' | 'Project' | 'Distinct' | 'Graph':
            yield from _extract_nanopublication_uris(algebra['p'])

        case 'BGP':
            for retractor, retracts, retractee in algebra['triples']:
                if retracts == URIRef(
                    'https://purl.org/nanopub/x/retracts',
                ) and isinstance(retractor, Variable):
                    yield retractee

        case 'LeftJoin' | 'Join':
            yield from _extract_nanopublication_uris(algebra['p1'])
            yield from _extract_nanopublication_uris(algebra['p2'])

        case 'Filter' | 'OrderBy':
            return

        case unknown_name:
            raise ValueError(
                f'Unknown algebra name: {unknown_name}, '
                f'content: {algebra}',
            )


def apply_redirect(source: URIRef) -> URIRef:
    """Rewrite the URL."""
    for pattern, destination in REDIRECTS.items():
        if source.startswith(pattern):
            return destination

    return source


@dataclasses.dataclass(frozen=True)
class GlobalSPARQLProcessor(Processor):  # noqa: WPS338, WPS214
    """
    Execute SPARQL queries against the whole Linked Data Web, or The Cyberspace.

    When running the queries, we will try to find and to import pieces of LD
    which can be relevant to the query we are executing.
    """

    graph: ConjunctiveGraph
    inference_lock: Lock = dataclasses.field(default_factory=Lock)
    logger: Any = loguru.logger

    def __post_init__(self):
        """Note that we do not presently need OWL inference."""
        self.graph.last_not_inferred_source = None
        self.graph._indices_loaded = False

    def _infer_with_sparql(self):
        """
        Infer triples with SPARQL rules.

        FIXME:
          * Code these rules into SHACL or some other RDF based syntax;
          * Make them available at iolanta.tech/visualizations/ and indexed.
        """
        inference = Path(__file__).parent / 'inference'

        file_names = {
            'wikibase-claim.sparql': URIRef('local:inference-wikibase-claim'),
            'wikibase-statement-property.sparql': URIRef(
                'local:inference-statement-property',
            ),
        }

        for file_name, graph_name in file_names.items():
            start_time = time.time()
            self.graph.update(
                update_object=(inference / file_name).read_text(),
            )
            triple_count = len(self.graph.get_context(graph_name))
            duration = datetime.timedelta(seconds=time.time() - start_time)
            self.logger.info(
                f'{file_name}: {triple_count} triple(s), '
                f'inferred at {duration}',
            )

    def maybe_apply_inference(self):
        """Apply global OWL RL inference if necessary."""
        if not REASONING_ENABLED:
            return

        if self.graph.last_not_inferred_source is None:
            return

        with self.inference_lock:
            self._infer_with_sparql()
            self._infer_with_owl_rl()
            self.logger.info('Inference @ cyberspace: complete.')

            self.graph.last_not_inferred_source = None

    def _infer_with_owl_rl(self):
        if not OWL_REASONING_ENABLED:
            return

        reasoner = reasonable.PyReasoner()
        reasoner.from_graph(self.graph)
        inferred_triples = reasoner.reason()
        inference_graph_name = BNode('_:inference')
        inferred_quads = [
            (*triple, inference_graph_name)
            for triple in inferred_triples
        ]
        self.graph.addN(inferred_quads)

    def _maybe_load_indices(self):
        if not self.graph._indices_loaded:
            for index in INDICES:
                self.load(index)

            self.graph._indices_loaded = True

    def query(   # noqa: WPS211, WPS210, WPS231, C901
        self,
        strOrQuery,
        initBindings=None,
        initNs=None,
        base=None,
        DEBUG=False,
    ):
        """
        Evaluate a query with the given initial bindings, and initial
        namespaces. The given base is used to resolve relative URIs in
        the query and will be overridden by any BASE given in the query.
        """
        self._maybe_load_indices()

        initBindings = initBindings or {}
        initNs = initNs or {}

        if isinstance(strOrQuery, Query):
            query = strOrQuery

        else:
            parse_tree = parseQuery(strOrQuery)
            query = translateQuery(parse_tree, base, initNs)

            self.load_retracting_nanopublications_by_query(
                query=query,
                bindings=initBindings,
                base=base,
                namespaces=initNs,
            )

            query, urls = extract_mentioned_urls_from_query(
                query=query,
                bindings=initBindings,
                base=base,
                namespaces=initNs,
            )

            for url in urls:
                try:
                    self.load(url)
                except Exception as err:
                    self.logger.error(f'Failed to load {url}: {err}', url, err)

        self.maybe_apply_inference()

        is_anything_loaded = True
        while is_anything_loaded:
            is_anything_loaded = False

            query_result = evalQuery(self.graph, query, initBindings, base)

            bindings = list(query_result['bindings'])
            for row in bindings:
                for _, maybe_iri in row.items():
                    if (
                        isinstance(maybe_iri, URIRef)
                        and isinstance(self.load(maybe_iri), Loaded)
                    ):
                        is_anything_loaded = True   # noqa: WPS220
                        self.logger.info(   # noqa: WPS220
                            'Newly loaded: {uri}',
                            uri=maybe_iri,
                        )

        query_result['bindings'] = bindings
        return query_result

    def _is_loaded(self, uri: URIRef) -> bool:
        """Find out if this URI in the graph already."""
        return funcy.first(
            self.graph.quads((
                uri,
                IOLANTA['last-loaded-time'],
                None,
                URIRef('iolanta://_meta'),
            )),
        ) is not None

    def _mark_as_loaded(self, uri: URIRef):
        self.graph.add((
            uri,
            IOLANTA['last-loaded-time'],
            Literal(datetime.datetime.now()),
            URIRef('iolanta://_meta'),
        ))

    def _follow_is_visualized_with_links(self, uri: URIRef):
        """Follow `dcterms:isReferencedBy` links."""
        self.logger.info(f'Following links for {uri}…')
        triples = self.graph.triples((uri, DCTERMS.isReferencedBy, None))
        for _, _, visualization in triples:
            if isinstance(visualization, URIRef):
                self.load(visualization)
        self.logger.info('Links followed.')

    def load(   # noqa: C901, WPS210, WPS212, WPS213, WPS231
        self,
        source: URIRef,
    ) -> LoadResult:
        """
        Try to load LD denoted by the given `source`.

        TODO This function is too big, we have to refactor it.
        """
        url = URL(source)

        if url.scheme in {'file', 'python', 'local', 'urn', 'doi'}:
            # FIXME temporary fix. `yaml-ld` doesn't read `context.*` files and
            #   fails.
            return Skipped()

        if url.fragment:
            # Fragment on an HTML page resolves to that same page. Let us remove
            # this ambiguity, then.
            # TODO: It works differently for JSON-LD documents AFAIK. Need to
            #   double check that.
            url = url.with_fragment(None)
            source = URIRef(str(f'{url}#'))

        self._follow_is_visualized_with_links(source)

        new_source = apply_redirect(source)
        if new_source != source:
            self.logger.info(
                'Rewriting: {source} → {new_source}',
                source=source,
                new_source=new_source,
            )
            return self.load(new_source)

        source_uri = normalize_term(source)
        if self._is_loaded(source_uri):
            return Skipped()
        else:
            self.logger.info(f'{source_uri} is not loaded yet')

        # FIXME This is definitely inefficient. However, python-yaml-ld caches
        #   the document, so the performance overhead is not super high.
        try:
            _resolved_source = yaml_ld.load_document(source)['documentUrl']
        except NotFound as not_found:
            self.logger.info(f'{not_found.path} | 404 Not Found')
            namespaces = [RDF, RDFS, OWL, FOAF, DC, VANN]

            for namespace in namespaces:
                if not_found.path.startswith(str(namespace)):
                    self.load(URIRef(namespace))
                    self.logger.info(
                        'Redirecting %s → namespace %s',
                        not_found.path,
                        namespace,
                    )
                    return Loaded()

            self.logger.info(
                '{path} | Cannot find a matching namespace',
                path=not_found.path,
            )

            self.graph.add((
                source_uri,
                RDF.type,
                IOLANTA.Graph,
            ))

            self.graph.add((
                source_uri,
                RDF.type,
                IOLANTA['not-found'],
                source_uri,
            ))

            self.graph.add((
                IOLANTA.Graph,
                RDF.type,
                RDFS.Class,
            ))

            self._mark_as_loaded(source_uri)

            return Loaded()

        except Exception as err:
            self.logger.info(f'{source} | Failed: {err}')

            self.graph.add((
                source_uri,
                RDF.type,
                IOLANTA.Graph,
            ))

            self.graph.add((
                URIRef(source),
                RDF.type,
                IOLANTA['failed'],
                source_uri,
            ))

            self.graph.add((
                IOLANTA.Graph,
                RDF.type,
                RDFS.Class,
            ))

            self._mark_as_loaded(source_uri)

            return Loaded()

        if _resolved_source:
            _resolved_source_uri_ref = URIRef(_resolved_source)
            if _resolved_source_uri_ref != URIRef(source):
                self.graph.add((
                    source_uri,
                    IOLANTA['redirects-to'],
                    _resolved_source_uri_ref,
                ))
                source = _resolved_source

        self.graph.add((
            source_uri,
            RDF.type,
            IOLANTA.Graph,
        ))

        self.graph.add((
            IOLANTA.Graph,
            RDF.type,
            RDFS.Class,
        ))

        self._mark_as_loaded(source_uri)

        try:  # noqa: WPS225
            ld_rdf = yaml_ld.to_rdf(source)
        except ConnectionError as name_resolution_error:
            self.logger.info(
                '%s | name resolution error: %s',
                source,
                str(name_resolution_error),
            )
            return Loaded()
        except ParserNotFound as parser_not_found:
            self.logger.info(f'{source} | {parser_not_found}')
            return Loaded()
        except YAMLLDError as yaml_ld_error:
            self.logger.error(f'{source} | {yaml_ld_error}')
            return Loaded()

        try:
            quads = list(
                parse_quads(
                    quads_document=ld_rdf,
                    graph=source_uri,
                    blank_node_prefix=str(source),
                ),
            )
        except UnresolvedIRI as err:
            raise dataclasses.replace(
                err,
                context=None,
                iri=source,
            )

        if not quads:
            self.logger.warning('{source} | No data found', source=source)
            self.graph.addN([(
                source_uri,
                RDF.type,
                IOLANTA.Graph,
                source_uri,
            )])
            return Loaded()

        quad_tuples = [
            tuple([
                normalize_term(term) for term in quad.as_tuple()
            ])
            for quad in quads
        ]

        self.graph.addN(quad_tuples)
        self.graph.last_not_inferred_source = source

        into_graphs = ', '.join({
            quad.graph
            for quad in quads
        })
        self.logger.info(
            f'{source} | loaded successfully into graphs: {into_graphs}',
        )
        return Loaded()

    def resolve_term(self, term: Node, bindings: dict[str, Node]):
        """Resolve triple elements against initial variable bindings."""
        if isinstance(term, Variable):
            return bindings.get(
                str(term),
                term,
            )

        return term

    def load_retracting_nanopublications_by_query(    # noqa: WPS231
        self,
        query: Query,
        bindings: dict[str, Node],
        base: str | None,
        namespaces: dict[str, Namespace],
    ):
        """
        If the query requires information about retracting nanopubs, load them.

        FIXME: This function presently does nothing because `nanopub` library
          has problems: https://github.com/fair-workflows/nanopub/issues/168

        TODO: Generalize this mechanism to allow for plugins which analyze
          SPARQL queries and load information into the graph based on their
          content.
        """
        nanopublications = list(
            resolve_variables(
                terms=_extract_nanopublication_uris(query.algebra),
                bindings=bindings,
            ),
        )

        for nanopublication in nanopublications:
            retractions = find_retractions_for(nanopublication)
            for retraction in retractions:
                self.load(retraction)



import locale
import logging
from pathlib import Path
from typing import Annotated

import loguru
import platformdirs
from documented import DocumentedError
from rdflib import Literal, URIRef
from rich.console import Console
from rich.markdown import Markdown
from typer import Argument, Exit, Option, Typer
from yarl import URL

from iolanta.cli.models import LogLevel
from iolanta.iolanta import Iolanta
from iolanta.models import NotLiteralNode

DEFAULT_LANGUAGE = locale.getlocale()[0].split('_')[0]


console = Console()


def construct_app() -> Typer:
    """
    Construct Typer app.

    FIXME: Remove this function, just create the app on module level.
    """
    iolanta = Iolanta()

    return Typer(
        no_args_is_help=True,
        context_settings={
            'obj': iolanta,
        },
    )


app = construct_app()


def string_to_node(name: str) -> NotLiteralNode:
    """
    Parse a string into a node identifier.

    String might be:
      * a URL,
      * or a local disk path.
    """
    url = URL(name)
    if url.scheme:
        return URIRef(name)

    path = Path(name).absolute()
    return URIRef(f'file://{path}')


@app.command(name='browse')
def render_command(   # noqa: WPS231, WPS238, WPS210, C901
    url: Annotated[str, Argument()],
    as_datatype: Annotated[
        str, Option(
            '--as',
        ),
    ] = 'https://iolanta.tech/cli/interactive',
    language: Annotated[
        str, Option(
            help='Data language to prefer.',
        ),
    ] = DEFAULT_LANGUAGE,
    log_level: LogLevel = LogLevel.ERROR,
):
    """Render a given URL."""
    level = {
        LogLevel.DEBUG: logging.DEBUG,
        LogLevel.INFO: logging.INFO,
        LogLevel.WARNING: logging.WARNING,
        LogLevel.ERROR: logging.ERROR,
    }[log_level]

    log_file_path = platformdirs.user_log_path(
        'iolanta',
        ensure_exists=True,
    ) / 'iolanta.log'

    logger = loguru.logger
    logger.add(
        log_file_path,
        level=level,
        format='{time} {level} {message}',
        enqueue=True,
    )

    node_url = URL(url)
    if node_url.scheme:
        node = URIRef(url)

        iolanta: Iolanta = Iolanta(
            language=Literal(language),
            logger=logger,
        )
    else:
        path = Path(url).absolute()
        node = URIRef(f'file://{path}')
        iolanta: Iolanta = Iolanta(
            language=Literal(language),
            logger=logger,
            project_root=path,
        )

    try:
        renderable, stack = iolanta.render(
            node=URIRef(node),
            as_datatype=URIRef(as_datatype),
        )

    except DocumentedError as documented_error:
        if iolanta.logger.level in {logging.DEBUG, logging.INFO}:
            raise

        console.print(
            Markdown(
                str(documented_error),
                justify='left',
            ),
        )
        raise Exit(1)

    except Exception as err:
        if level in {logging.DEBUG, logging.INFO}:
            raise

        console.print(str(err))
        raise Exit(1)

    else:
        Console().print(renderable)



from dataclasses import dataclass, field

from rdflib import URIRef
from rich.table import Table

from iolanta.facets.cli import Renderable, RichFacet
from iolanta.namespaces import IOLANTA, OWL


@dataclass
class Record(RichFacet):
    skipped_properties: set[URIRef] = field(
        default_factory=lambda: {
            OWL.sameAs,
        },
    )

    def show(self) -> Renderable:
        return "RECORD"
        rows = self.stored_query('record.sparql', node=self.iri)

        caption = self.render(self.iri, as_datatype=IOLANTA['cli/title'])

        table = Table(
            show_header=False,
            title=caption,
        )

        rows = [
            row for row in rows
            if row['property'] not in self.skipped_properties
        ]

        if not rows:
            return caption

        for row in rows:
            rendered_property = self.render(
                row['property'],
                as_datatype=IOLANTA['cli/record/property'],
            )
            rendered_value = self.render(
                row['value'],
                as_datatype=IOLANTA['cli/record/value'],
            )
            table.add_row(rendered_property, rendered_value)

        return table



from dataclasses import dataclass
from functools import cached_property
from graphlib import TopologicalSorter
from typing import Iterable, List, TypedDict

import funcy
from rdflib.term import Literal, Node, URIRef
from yarl import URL

from iolanta.facets.errors import FacetNotFound
from iolanta.models import NotLiteralNode


class FoundRow(TypedDict):
    """Facet and datatype to render an IRI."""

    facet: NotLiteralNode
    as_datatype: NotLiteralNode


def reorder_rows_by_facet_preferences(   # noqa: WPS214, WPS210
    rows: list[FoundRow],
    ordering: set[tuple[URIRef, URIRef]],
) -> list[FoundRow]:
    """
    Apply a partial ordering to given rows.

    Preserve existing ordering.

    Note: I must confess I did not dive into this logic, it was what ChatGPT
    had suggested to me and it seemed working.

    To be fair, this should be understood and unit tested.
    """
    # Map each facet to its index in the original list for stable ordering
    row_index_map = {row['facet']: index for index, row in enumerate(rows)}

    # Initialize the topological sorter
    ts = TopologicalSorter()

    # Add edges to the graph based on ordering
    for before, after in ordering:
        if before in row_index_map and after in row_index_map:
            ts.add(after, before)

    # Compute the topological order
    sorted_facets = list(ts.static_order())

    # Append any remaining facets not in the ordering, preserving their original
    # order
    remaining_facets = [
        facet
        for facet in row_index_map
        if facet not in sorted_facets
    ]

    # Combine sorted and remaining facets
    final_order = sorted_facets + remaining_facets

    # Map back to rows and sort them
    return sorted(
        rows,
        key=lambda row: final_order.index(row['facet']),
    )


@dataclass
class FacetFinder:   # noqa: WPS214
    """Engine to find facets for a given node."""

    iolanta: 'iolanta.Iolanta'    # type: ignore
    node: Node
    as_datatype: NotLiteralNode

    @cached_property
    def row_sorter_by_output_datatype(self):
        def _sorter(row) -> int:
            return 0

        return _sorter

    def by_datatype(self) -> Iterable[FoundRow]:
        if not isinstance(self.node, Literal):
            return []

        if (data_type := self.node.datatype) is None:
            return []

        rows = self.iolanta.query(   # noqa: WPS462
            """
            SELECT ?output_datatype ?facet WHERE {
                $data_type iolanta:hasDatatypeFacet ?facet .
                ?facet iolanta:outputs ?output_datatype .
            }
            """,
            data_type=data_type,
        )

        rows = [row for row in rows if row['output_datatype'] == self.as_datatype]

        return sorted(
            rows,
            key=self.row_sorter_by_output_datatype,
        )

    def by_prefix(self) -> Iterable[FoundRow]:
        """Determine facet by URL prefix.

        TODO fix this to allow arbitrary prefixes.
        """
        scheme = URL(self.node).scheme
        if scheme != 'urn':
            return []

        if not isinstance(self.node, URIRef):
            return []

        rows = self.iolanta.query(   # noqa: WPS462
            """
            SELECT ?output_datatype ?facet WHERE {
                $prefix iolanta:hasFacetByPrefix ?facet .
                ?facet iolanta:outputs ?output_datatype .
            }
            """,
            prefix=URIRef(f'{scheme}:'),
        )

        rows = [row for row in rows if row['output_datatype'] == self.as_datatype]

        return sorted(
            rows,
            key=self.row_sorter_by_output_datatype,
        )

    def by_facet(self) -> List[FoundRow]:
        """Find a facet directly attached to the node."""
        if isinstance(self.node, Literal):
            return []

        rows = self.iolanta.query(   # noqa: WPS462
            """
            SELECT ?output_datatype ?facet WHERE {
                $node iolanta:facet ?facet .
                ?facet iolanta:outputs ?output_datatype .
            }
            """,
            node=self.node,
        )

        # FIXME This is probably suboptimal, why don't we use `IN output_datatypes`?
        rows = [row for row in rows if row['output_datatype'] == self.as_datatype]

        return sorted(
            rows,
            key=self.row_sorter_by_output_datatype,
        )

    def _classes_for_node(self, node: NotLiteralNode) -> list[NotLiteralNode]:
        rows = self.iolanta.query(  # noqa: WPS462
            """
            SELECT DISTINCT ?class WHERE {
                $node
                    rdf:type / (owl:equivalentClass|^owl:equivalentClass)*
                    ?class .
            }
            """,
            node=node,
        )

        return funcy.lpluck('class', rows)

    def by_instance_facet(self) -> Iterable[FoundRow]:
        """Find facet by classes the IRI belongs to."""
        classes = self._classes_for_node(self.node)
        if not classes:
            return []

        formatted_classes = ', '.join(f'<{iri}>' for iri in classes)

        rows: list[FoundRow] = self.iolanta.query(   # noqa: WPS462
            """
            SELECT ?output_datatype ?facet WHERE {
                ?class iolanta:hasInstanceFacet ?facet .
                ?facet iolanta:outputs ?output_datatype .
                FILTER(?class IN (%s)) .
            }
            """ % formatted_classes,
            classes=classes,
            output_datatype=self.as_datatype,
        )

        return sorted(
            rows,
            key=self.row_sorter_by_output_datatype,
        )

    def by_output_datatype_default_facet(self) -> Iterable[FoundRow]:
        """Find facet based on output_datatype only."""
        rows = self.iolanta.query(
            '''
            SELECT ?facet ?output_datatype WHERE {
              $output_datatype iolanta:hasDefaultFacet ?facet .
            }
            ''',
            output_datatype=self.as_datatype,
        )

        return [
            FoundRow(facet=row['facet'], output_datatype=row['output_datatype'])
            for row in rows
        ]

    def retrieve_facets_preference_ordering(self) -> set[tuple[URIRef, URIRef]]:
        """
        Construct partial ordering on the set of facets.

        In each pair, the first element is preferred over the latter.
        """
        rows = self.iolanta.query(  # noqa: WPS462
            """
            SELECT ?preferred ?over WHERE {
                ?preferred iolanta:is-preferred-over ?over
            }
            """,
        )

        return {
            (row['preferred'], row['over'])
            for row in rows
        }

    def choices(self) -> list[FoundRow]:
        """Return all suitable facets."""
        rows = list(self._found_facets())

        if len(rows) == 1:
            # Nothing to order.
            return rows

        ordering = self.retrieve_facets_preference_ordering()
        return reorder_rows_by_facet_preferences(rows, ordering)

    @property
    def facet_and_output_datatype(self) -> FoundRow:
        """Choose the best facet for the IRI."""
        if choice := funcy.first(self.choices()):
            return choice

        raise FacetNotFound(
            node=self.node,
            as_datatype=self.as_datatype,
            node_types=[],
        )

    def _found_facets(self) -> Iterable[FoundRow]:
        """Compose a stream of all possible facet choices."""
        yield from self.by_prefix()
        yield from self.by_datatype()
        yield from self.by_facet()
        yield from self.by_instance_facet()
        yield from self.by_output_datatype_default_facet()



import funcy

from iolanta import Facet
from iolanta.namespaces import DATATYPES


class FOAFPersonTitle(Facet[str]):
    """Show title for a foaf:Person object."""

    def show(self) -> str:
        """Render full name of a person."""
        row = funcy.first(self.stored_query('names.sparql', person=self.iri))

        if row is None:
            # Render default title.
            return self.render(
                self.iri,
                as_datatype=DATATYPES['fallback-title'],
            )

        family_name = row['family_name']
        given_name = row['given_name']

        return f'{given_name} {family_name}'



import sys
from typing import List, Type, TypeVar

PluginClass = TypeVar('PluginClass')


if sys.version_info < (3, 10):    # pragma: no cover
    from importlib_metadata import entry_points
else:    # pragma: no cover
    from importlib.metadata import entry_points


def plugins(group_name: str) -> List[PluginClass]:
    """List of plugin classes by group name."""
    return [
        entry_point.load()  # type: ignore
        for entry_point in entry_points(group=group_name)  # type: ignore
    ]



import logging

from classes import typeclass
from rdflib import Graph, URIRef
from rdflib.term import Node

from iolanta.models import ComputedQName

logger = logging.getLogger(__name__)


@typeclass
def node_to_qname(node: Node, graph: Graph):
    """Convert a node to a QName."""


@node_to_qname.instance(object)
def _object_to_qname(node: URIRef, graph: Graph):
    return node


@node_to_qname.instance(URIRef)
def _uriref_to_qname(node: URIRef, graph: Graph):
    try:
        qname = ComputedQName(*graph.compute_qname(node))   # type: ignore

    except ValueError:
        return node

    except NameError as err:
        logger.exception(f'NameError! On: {node} Error: {err}')
        return node

    if qname.namespace_name.startswith('ns'):
        return node

    return qname



import json
import sys

from classes import typeclass
from rdflib import Graph

from ldflex.ldflex import QueryResult, SelectResult


@typeclass
def print_json(query_result: QueryResult) -> None:
    """Print query result as JSON."""


@print_json.instance(SelectResult)
def _print_select_as_json(select_result: SelectResult):
    sys.stdout.write(json.dumps(select_result, indent=2, default=str))


@print_json.instance(Graph)
def _print_construct_as_json(graph: Graph):
    fieldnames = ('subject', 'predicate', 'object')
    sys.stdout.write(
        json.dumps(
            [
                dict(zip(fieldnames, triple))
                for triple in graph
            ],
            indent=2,
            default=str,
        ),
    )



import dataclasses
import hashlib
from typing import Iterable

from rdflib import BNode, Literal, URIRef
from rdflib.term import Node

from iolanta.errors import UnresolvedIRI
from iolanta.models import Quad
from iolanta.namespaces import IOLANTA, RDF
from iolanta.parsers.errors import SpaceInProperty


def parse_term(   # noqa: C901
    term,
    blank_node_prefix,
) -> Node:
    """Parse N-Quads term into a Quad."""
    if term is None:
        raise SpaceInProperty()

    term_type = term['type']
    term_value = term['value']

    if term_type == 'IRI':
        return URIRef(term_value)

    if term_type == 'literal':
        language = term.get('language')

        if datatype := term.get('datatype'):
            datatype = URIRef(datatype)

        if language and datatype:
            datatype = None

        return Literal(
            term_value,
            datatype=datatype,
            lang=language,
        )

    if term_type == 'blank node':
        return BNode(
            value=term_value.replace('_:', f'{blank_node_prefix}_'),
        )

    raise ValueError(f'Unknown term: {term}')


def parse_quads(
    quads_document,
    graph: URIRef,
    blank_node_prefix: str = '',
) -> Iterable[Quad]:
    """Parse an N-Quads output into a Quads stream."""
    blank_node_prefix = hashlib.md5(  # noqa: S324
        blank_node_prefix.encode(),
    ).hexdigest()
    blank_node_prefix = f'_:{blank_node_prefix}'

    for graph_name, quads in quads_document.items():
        if graph_name == '@default':
            graph_name = graph   # noqa: WPS440

        else:
            graph_name = URIRef(graph_name)

            yield Quad(
                graph,
                IOLANTA['has-sub-graph'],
                graph_name,
                graph,
            )

        yield Quad(
            graph_name,
            RDF.type,
            IOLANTA.Graph,
            IOLANTA.meta,
        )

        for quad in quads:
            try:
                yield Quad(
                    subject=parse_term(quad['subject'], blank_node_prefix),
                    predicate=parse_term(quad['predicate'], blank_node_prefix),
                    object=parse_term(quad['object'], blank_node_prefix),
                    graph=graph_name,
                )
            except SpaceInProperty as err:
                raise dataclasses.replace(
                    err,
                    iri=graph,
                )


def raise_if_term_is_qname(term_value: str):
    """Raise an error if a QName is provided instead of a full IRI."""
    prefix, etc = term_value.split(':', 1)

    if etc.startswith('/'):
        return

    if prefix in {'local', 'templates', 'urn'}:
        return

    raise UnresolvedIRI(
        iri=term_value,
        prefix=prefix,
    )



import logging

from textual.app import ReturnType
from textual.logging import TextualHandler

from iolanta.facets.facet import Facet, FacetOutput
from iolanta.facets.textual_browser.app import IolantaBrowser


class TextualBrowserFacet(Facet[ReturnType | None]):
    """Textual browser."""

    def show(self) -> ReturnType | None:
        """Render the Iolanta browser Textual app."""
        logging.basicConfig(
            level=logging.DEBUG,
            handlers=[TextualHandler()],
            force=True,
        )

        app = IolantaBrowser(
            iolanta=self.iolanta,
            iri=self.iri,
        )
        try:
            app.run()
        except Exception:
            logging.exception("Unhandled exception.")



from rdflib import Literal, URIRef
from rich.style import Style
from rich.text import Text

from iolanta.facets.facet import Facet
from iolanta.namespaces import DATATYPES


class TextualLinkFacet(Facet[str | Text]):
    """Render a link within a Textual app."""

    def show(self) -> str | Text:
        """Render the link, or literal text, whatever."""
        if isinstance(self.iri, Literal):
            return f'[b grey37]{self.iri}[/b grey37]'

        label = self.render(
            self.iri,
            as_datatype=DATATYPES.title,
        )

        invocation = f"app.goto('{self.iri}')"
        return f'[@click="{invocation}"]{label}[/]'



import functools
import threading
import uuid
from dataclasses import dataclass
from typing import Any

import watchfiles
from rdflib import BNode, URIRef
from textual.widgets import ContentSwitcher, RichLog
from textual.worker import Worker, WorkerState

from iolanta.facets.errors import FacetError, FacetNotFound
from iolanta.facets.locator import FacetFinder
from iolanta.facets.textual_browser.history import NavigationHistory
from iolanta.facets.textual_browser.home import Home
from iolanta.facets.textual_browser.location import Location
from iolanta.facets.textual_browser.models import FlipOption
from iolanta.facets.textual_browser.page import Page
from iolanta.iolanta import Iolanta
from iolanta.models import NotLiteralNode
from iolanta.namespaces import DATATYPES
from iolanta.widgets.mixin import IolantaWidgetMixin


@dataclass
class RenderResult:
    """
    We asked a thread to render something for us.

    This is what did we get back.
    """

    iri: NotLiteralNode
    renderable: Any
    flip_options: list[FlipOption]
    facet_iri: URIRef
    is_reload: bool


class PageSwitcher(IolantaWidgetMixin, ContentSwitcher):  # noqa: WPS214
    """
    Container for open pages.

    Able to navigate among them while traversing the history.
    """

    BINDINGS = [  # noqa: WPS115
        ('alt+left', 'back', 'Back'),
        ('alt+right', 'forward', 'Fwd'),
        ('f5', 'reload', '🔄 Reload'),
        ('escape', 'abort', '🛑 Abort'),
        ('f12', 'console', 'Console'),
    ]

    def __init__(self):
        """Set Home as first tab."""
        super().__init__(id='page_switcher', initial='home')
        self.stop_file_watcher_event = threading.Event()

    def action_console(self):
        """Open dev console."""
        console_switcher = self.app.query_one(ConsoleSwitcher)
        console_switcher.current = 'console'
        console_switcher.query_one(DevConsole).focus()

    @functools.cached_property
    def history(self) -> NavigationHistory[Location]:
        """Cached navigation history."""
        return NavigationHistory[Location]()

    def compose(self):
        """Home is the first page to open."""
        yield Home(id='home')

    def on_mount(self):
        """Navigate to the initial page."""
        self.action_goto(self.app.iri)
        if self.iolanta.project_root:
            self.run_worker(
                self._watch_files,
                thread=True,
            )

    def on_unmount(self) -> None:
        """Stop watching files."""
        self.stop_file_watcher_event.set()

    def _watch_files(self):
        for _ in watchfiles.watch(   # noqa: WPS352
            self.iolanta.project_root,
            stop_event=self.stop_file_watcher_event,
        ):
            self.app.call_from_thread(self.action_reload)

    def render_iri(   # noqa: WPS210
        self,
        destination: NotLiteralNode,
        facet_iri: URIRef | None,
        is_reload: bool,
    ) -> RenderResult:
        """Render an IRI in a thread."""
        self.iri = destination
        iolanta: Iolanta = self.iolanta

        as_datatype = URIRef('https://iolanta.tech/cli/textual')
        choices = FacetFinder(
            iolanta=self.iolanta,
            node=destination,
            as_datatype=as_datatype,
        ).choices()

        if not choices:
            raise FacetNotFound(
                node=self.iri,
                as_datatype=as_datatype,
                node_types=[],
            )

        if facet_iri is None:
            facet_iri = choices[0]['facet']

        other_facets = [
            choice['facet']
            for choice in choices
            if choice['facet'] != facet_iri
        ]
        flip_options = [
            FlipOption(
                facet_iri=facet,
                title=self.app.call_from_thread(
                    self.iolanta.render,
                    facet,
                    as_datatype=DATATYPES.title,
                )[0],
            )
            for facet in other_facets
        ]

        facet_class = iolanta.facet_resolver[facet_iri]

        facet = facet_class(
            iri=self.iri,
            iolanta=iolanta,
            as_datatype=URIRef('https://iolanta.tech/cli/textual'),
        )

        try:
            renderable = facet.show()

        except Exception as err:
            raise FacetError(
                node=self.iri,
                facet_iri=facet_iri,
                error=err,
            ) from err

        return RenderResult(
            iri=destination,
            renderable=renderable,
            flip_options=flip_options,
            facet_iri=facet_iri,
            is_reload=is_reload,
        )

    def on_worker_state_changed(   # noqa: WPS210
        self,
        event: Worker.StateChanged,
    ):
        """Render a page as soon as it is ready."""
        match event.state:
            case WorkerState.SUCCESS:
                render_result: RenderResult = event.worker.result

                if render_result.is_reload:
                    # We are reloading the current page.
                    current_page = self.query_one(f'#{self.current}', Page)
                    current_page.remove_children()
                    current_page.mount(render_result.renderable)

                    # FIXME: This does not actually change the flip options,
                    #   but maybe that's okay
                    current_page.flip_options = render_result.flip_options

                else:
                    # We are navigating to a new page.
                    page_uid = uuid.uuid4().hex
                    page_id = f'page_{page_uid}'
                    page = Page(
                        render_result.renderable,
                        iri=render_result.iri,
                        page_id=page_id,
                        flip_options=render_result.flip_options,
                    )
                    self.mount(page)
                    self.current = page_id
                    page.focus()
                    self.history.goto(
                        Location(
                            page_id,
                            url=render_result.iri,
                            facet_iri=render_result.facet_iri,
                        ),
                    )
                    self.app.sub_title = render_result.iri

            case WorkerState.ERROR:
                raise ValueError(event)

    @property
    def is_loading(self) -> bool:
        """Determine if the app is presently loading something."""
        for worker in self.workers:
            if worker.name == 'render_iri':
                return True

        return False

    def action_reload(self):
        """Reset Iolanta graph and re-render current view."""
        self.iolanta.reset()

        self.run_worker(
            functools.partial(
                self.render_iri,
                destination=self.history.current.url,
                facet_iri=self.history.current.facet_iri,
                is_reload=True,
            ),
            thread=True,
            exclusive=True,
            name='render_iri',
        )
        self.refresh_bindings()

    def action_abort(self):
        """Abort loading."""
        self.notify(
            'Aborted.',
            severity='warning',
        )

        for worker in self.workers:
            if worker.name == 'render_iri':
                worker.cancel()
                break

        self.refresh_bindings()

    def check_action(
        self,
        action: str,
        parameters: tuple[object, ...],   # noqa: WPS110
    ) -> bool | None:
        """Check if action is available."""
        is_loading = self.is_loading
        match action:
            case 'reload':
                return not is_loading
            case 'abort':
                return is_loading
            case 'back':
                return bool(self.history.past)
            case 'forward':
                return bool(self.history.future)

        return True

    def action_goto(
        self,
        destination: str,
        facet_iri: str | None = None,
    ):
        """Go to an IRI."""
        if destination.startswith('_:'):
            iri = BNode(destination)
        else:
            iri = URIRef(destination)

        self.run_worker(
            functools.partial(
                self.render_iri,
                destination=iri,
                facet_iri=facet_iri and URIRef(facet_iri),
                is_reload=False,
            ),
            thread=True,
            exclusive=True,
            name='render_iri',
        )
        self.refresh_bindings()

    def action_back(self):
        """Go backward."""
        self.current = self.history.back().page_id
        if page := self.visible_content:
            page.focus()

    def action_forward(self):
        """Go forward."""
        self.current = self.history.forward().page_id
        self.focus()
        if page := self.visible_content:
            page.focus()


class ConsoleSwitcher(ContentSwitcher):
    """Switch between page switcher and dev console."""

    def __init__(self):
        """Specify initial params."""
        super().__init__(
            id='console_switcher',
            initial='page_switcher',
        )

    def compose(self):
        """Compose two tabs."""
        yield PageSwitcher()
        yield DevConsole()


class DevConsole(RichLog):
    """Development console."""

    BINDINGS = [
        ('f12,escape', 'close', 'Close Console'),
    ]

    def __init__(self):
        """Set default props for console."""
        super().__init__(highlight=True, markup=True, id='console')

    def action_close(self):
        """Close the dev console."""
        console_switcher = self.app.query_one(ConsoleSwitcher)
        console_switcher.current = 'page_switcher'

        page_switcher = console_switcher.query_one(PageSwitcher)
        page_switcher.visible_content.focus()



from typing import ClassVar

import funcy
from rdflib import URIRef
from rdflib.term import BNode, Literal, Node
from textual.app import ComposeResult, RenderResult
from textual.binding import Binding, BindingType
from textual.containers import Vertical
from textual.widget import Widget
from textual.widgets import Label

from iolanta.facets.facet import Facet
from iolanta.facets.textual_default.facets import PropertyName
from iolanta.facets.textual_default.triple_uri_ref import TripleURIRef
from iolanta.facets.textual_default.widgets import PropertyRow, Title
from iolanta.models import NotLiteralNode, Triple
from iolanta.namespaces import LOCAL, RDF


class RDFTermView(Widget, can_focus=True, inherit_bindings=False):
    """Display and navigate to an RDF term."""

    DEFAULT_CSS = """
    RDFTermView {
        width: auto;
        height: auto;
    }

    RDFTermView:hover {
        background: $boost;
    }

    RDFTermView:focus {
        background: darkslateblue;
    }
    """

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding('enter', 'goto', 'Goto'),
    ]

    def __init__(self, node: Node):
        """Initialize."""
        self.node = node
        super().__init__()

    def action_goto(self):
        """Navigate."""
        self.app.action_goto(self.node)

    def render(self) -> RenderResult:
        """Render the RDF term."""
        icon = {
            URIRef: '🔗',  # Represents a link or URL
            BNode: '⬜️',  # Represents a blank or undefined node
            Literal: '🔤',  # Represents text or a literal value
        }[type(self.node)]
        return f'{icon} {self.node}'


class ProvenanceView(Vertical):
    """Triple provenance page."""

    def __init__(
        self,
        triple: Triple,
        graphs: list[NotLiteralNode],
    ):
        """Initialize."""
        self.triple = triple
        self.graphs = graphs
        super().__init__()

    @property
    def graph_views(self):
        """Construct widgets to render the graph references."""
        return [RDFTermView(graph) for graph in self.graphs]

    def compose(self) -> ComposeResult:
        """Build page structure."""
        yield Title('Provenan©e for a triple')

        # TODO: Calculate QNames somehow.
        yield PropertyRow(
            PropertyName(RDF.subject, qname=str(RDF.subject)),
            RDFTermView(self.triple.subject),
        )

        yield PropertyRow(
            PropertyName(RDF.predicate, qname=str(RDF.predicate)),
            RDFTermView(self.triple.predicate),
        )

        yield PropertyRow(
            PropertyName(RDF.object, qname=str(RDF.object)),
            RDFTermView(self.triple.object),
        )

        yield Label('[b]Appearances[/b]')

        if self.graphs:
            yield from self.graph_views
        else:
            yield Label('None detected ☹')


class TextualProvenanceFacet(Facet[Widget]):
    """Facet for a triple."""

    def show(self) -> Widget:
        """Obtain & render provenance info."""
        uri = TripleURIRef(self.iri)
        triple = uri.as_triple()

        graphs = funcy.lpluck(
            'graph',
            self.stored_query(
                'graphs.sparql',
                subject=triple.subject,
                predicate=triple.predicate,
                object=triple.object,
            ),
        )

        if not graphs:
            exists = self.stored_query(
                'triples.sparql',
                subject=triple.subject,
                predicate=triple.predicate,
                object=triple.object,
            )

            if exists:
                graphs = [LOCAL.term('_inference')]

        return ProvenanceView(
            triple=triple,
            graphs=graphs,
        )



from iolanta.iolanta import Iolanta


class IolantaWidgetMixin:
    """Mixin for any Iolanta-powered Textual widget."""

    @property
    def iolanta(self) -> Iolanta:
        """Iolanta instance."""
        return self.app.iolanta    # type: ignore



import json
import re
from dataclasses import dataclass, field
from functools import reduce
from io import StringIO
from pathlib import Path
from typing import Iterable, List, Optional, TextIO, Type, Union

from documented import DocumentedError
from rdflib import URIRef
from rdflib.parser import URLInputSource
from requests import Response
from yarl import URL

from iolanta.context import merge
from iolanta.conversions import url_to_iri, url_to_path
from iolanta.loaders.base import Loader
from iolanta.loaders.errors import IsAContext, ParserNotFound
from iolanta.loaders.local_file import choose_parser_by_extension
from iolanta.models import LDContext, LDDocument, Quad
from iolanta.parsers.base import Parser
from iolanta.parsers.json import JSON
from iolanta.parsers.markdown import Markdown
from iolanta.parsers.yaml import YAML


@dataclass(frozen=True)
class HTTP(Loader[URL]):
    """
    Retrieve Linked Data from a file on the Web.
    """

    context: LDContext = field(default_factory=dict)

    def choose_parser_class(self, source: URL, response: Response):
        # FIXME hard code. Make this extensible.
        try:
            return choose_parser_by_extension(source)
        except ParserNotFound:
            content_type = response.headers['Content-Type']

            raise ValueError(f'Content type: {content_type}')

    def extract_alternate_url(
        self,
        source: URL,
        response: Response,
    ) -> URL | None:
        link = response.headers.get('Link')

        if link is None:
            return None

        match = re.match(
            r'<([^>]+)>; rel="alternate"; type="application/ld\+json"',
            link,
        )
        if match is None:
            return None

        return source / match.group(1)

    def as_jsonld_document(
        self,
        source: URL,
        iri: Optional[URIRef] = None,
    ) -> LDDocument:
        if iri is None:
            iri = url_to_iri(source)

        response = source.get()
        response.raise_for_status()
        alternate_url = self.extract_alternate_url(
            source=source,
            response=response,
        )
        if alternate_url is not None:
            return self.as_jsonld_document(
                source=alternate_url,
                iri=iri,
            )

        # `response.text` doesn't work.
        # Reasoning: https://stackoverflow.com/a/72621231/1245471
        response_as_file = StringIO(response.content.decode('utf-8'))

        parser_class: Type[Parser] = self.choose_parser_class(
            source=source,
            response=response,
        )
        try:
            document = parser_class().as_jsonld_document(response_as_file)
        except Exception:
            raise ValueError(response)

        if iri is not None and isinstance(document, dict):
            document.setdefault('@id', str(iri))

        return document

    def as_file(self, source: URL) -> TextIO:
        raise ValueError('!!!')

    def as_quad_stream(
        self,
        source: URL,
        iri: Optional[URIRef],
        root_loader: 'Loader[URL]',
    ) -> Iterable[Quad]:
        try:
            parser_class = self.choose_parser_class(source)
        except ParserNotFound:
            return []

        if iri is None:
            iri = url_to_iri(source)

        with source.open() as text_io:
            return parser_class().as_quad_stream(
                raw_data=text_io,
                iri=iri,
                context=self.context,
                root_loader=root_loader,
            )

    def find_context(self, source: str) -> LDContext:
        raise ValueError('??!!?')



import funcy

from iolanta import Facet


class IconFacet(Facet[str]):
    """Icon of an object."""

    def show(self) -> str:
        """Render icon of a thing."""
        row = funcy.first(
            self.query(   # noqa: WPS462
                """
                SELECT * WHERE {
                    $iri iolanta:icon ?icon .
                }
                """,
                iri=self.iri,
            ),
        )

        if row:
            return row['icon']

        return ''



import inspect
from abc import ABC
from dataclasses import dataclass, field
from logging import Logger
from pathlib import Path
from typing import List, Optional

from rdflib.term import Node
from typer import Typer

from iolanta.models import TripleTemplate


@dataclass
class Plugin(ABC):
    """Base Iolanta plugin."""

    iolanta: 'iolanta.Iolanta' = field(repr=False)

    @property
    def logger(self) -> Logger:
        return self.iolanta.logger

    @property
    def typer_app(self) -> Optional[Typer]:
        """Typer app for this plugin's CLI."""
        return None

    @property
    def files_directory(self) -> Path:
        return Path(inspect.getfile(self.__class__)).parent / 'data'

    @property
    def context_path(self) -> Optional[Path]:
        if (context_path := self.files_directory / 'context.yaml').is_file():
            return context_path

    @property
    def data_files(self):
        return self.files_directory

    def retrieve_triple(self, triple_template: TripleTemplate):
        """Save datasets which might describe the given node into project."""



import dataclasses
import itertools
import json
import uuid
from dataclasses import dataclass
from typing import Any, Iterable, Optional

from documented import DocumentedError
from pyld.jsonld import _resolved_context_cache  # noqa: WPS450
from pyld.jsonld import JsonLdError, expand, flatten, to_rdf  # noqa: WPS347
from rdflib import BNode, URIRef
from yarl import URL

from iolanta.errors import UnresolvedIRI
from iolanta.loaders import Loader
from iolanta.models import LDContext, LDDocument, NotLiteralNode, Quad
from iolanta.namespaces import IOLANTA, LOCAL, RDF
from iolanta.parse_quads import parse_quads
from iolanta.parsers.base import Parser, RawDataType


class DictParser(Parser[LDDocument]):
    """
    Old version of dict parser.

    FIXME: Remove this.
    """

    def as_jsonld_document(self, raw_data: LDDocument) -> LDDocument:
        """Do nothing."""
        return raw_data

    def as_quad_stream(
        self,
        raw_data: RawDataType,
        iri: Optional[NotLiteralNode],
        context: LDContext,
        root_loader: Loader,
    ) -> Iterable[Quad]:
        """Do nonsense."""
        # This helps avoid weird bugs when loading data.
        _resolved_context_cache.clear()

        document = raw_data

        if iri is None:
            uid = uuid.uuid4().hex
            iri = BNode(f'_:dict:{uid}')

        document = assign_key_if_not_present(
            document=document,
            key='iolanta:subjectOf',
            default_value={
                '$id': str(iri),
            },
        )

        try:
            document = expand(
                document,
                options={
                    'expandContext': context,
                    'documentLoader': root_loader,

                    # Explanation:
                    #   https://github.com/digitalbazaar/pyld/issues/143
                    'base': str(LOCAL),
                },
            )
        except (JsonLdError, KeyError, TypeError) as err:
            raise ExpandError(
                message=str(err),
                document=document,
                context=context,
                iri=iri,
                document_loader=root_loader,
            ) from err

        document = flatten(document)

        static_quads = [
            Quad(iri, RDF.type, IOLANTA.File, iri),
        ]

        try:
            parsed_quads = list(
                parse_quads(
                    quads_document=to_rdf(document),
                    # FIXME:
                    #   title: Can iri be None in a parser?
                    #   description: |
                    #     Does it make sense? If not, just change
                    #     the annotation.
                    graph=iri,  # type: ignore
                    blank_node_prefix=str(iri),
                ),
            )
        except UnresolvedIRI as err:
            raise dataclasses.replace(
                err,
                context=context,
                iri=iri,
            )

        return list(
            itertools.chain(
                parsed_quads,
                static_quads,
            ),
        )


def assign_key_if_not_present(  # type: ignore
    document: LDDocument,
    key: str,
    default_value: Any,
) -> LDDocument:
    """Add key to document if it does not exist yet."""
    if isinstance(document, dict):
        if document.get(key) is None:
            return {
                key: default_value,
                **document,
            }

        return document

    elif isinstance(document, list):
        return [
            assign_key_if_not_present(    # type: ignore
                document=sub_document,
                key=key,
                default_value=default_value,
            )
            for sub_document in document
        ]

    return document


@dataclass
class ExpandError(DocumentedError):
    """
    JSON-LD expand operation failed.

    IRI: {self.iri}

    Context: {self.formatted_context}

    Document: {self.formatted_data}

    Error: {self.message}

    Document Loader: {self.document_loader}
    """

    message: str
    document: LDDocument
    context: LDContext
    iri: Optional[URIRef]
    document_loader: Loader[URL]

    @property
    def formatted_data(self) -> str:
        """Format document for printing."""
        return json.dumps(self.document, indent=2, ensure_ascii=False)

    @property
    def formatted_context(self):
        """Format context for printing."""
        return json.dumps(self.context, indent=2, ensure_ascii=False)



import funcy
from rdflib import URIRef

from iolanta.facets.facet import Facet

PRIORITIES = [   # noqa: WPS407
    'dc_title',
    'schema_title',
    'schema_name',
    'rdfs_label',
    'foaf_name',
]


class TitleFacet(Facet[str]):
    """Title of an object."""

    def show(self) -> str:
        """Render title of a thing."""
        choices = self.stored_query(
            'title.sparql',
            iri=self.iri,
            language=self.iolanta.language,
        )

        if choices:
            row = funcy.first(choices)
            for alternative in PRIORITIES:
                if label := row.get(alternative):
                    return str(label)

        return self.render(
            self.iri,
            as_datatype=URIRef('https://iolanta.tech/qname'),
        )



import inspect
from dataclasses import dataclass, field
from functools import cached_property
from pathlib import Path
from typing import Any, Generic, Iterable, List, Optional, TypeVar, Union

from rdflib.term import BNode, Literal, Node, URIRef

from iolanta.models import NotLiteralNode, Triple, TripleTemplate
from iolanta.stack import Stack
from ldflex import LDFlex
from ldflex.ldflex import QueryResult, SPARQLQueryArgument

FacetOutput = TypeVar('FacetOutput')


@dataclass
class Facet(Generic[FacetOutput]):
    """Base facet class."""

    iri: NotLiteralNode
    iolanta: 'iolanta.Iolanta' = field(repr=False)
    as_datatype: Optional[NotLiteralNode] = None
    stack_children: List[Stack] = field(default_factory=list, repr=False)

    @property
    def stored_queries_path(self) -> Path:
        """Construct directory for stored queries for this facet."""
        return Path(inspect.getfile(self.__class__)).parent / 'sparql'

    @property
    def ldflex(self) -> LDFlex:
        """Extract LDFLex instance."""
        return self.iolanta.ldflex

    @cached_property
    def uriref(self) -> NotLiteralNode:
        """Format as URIRef."""
        if isinstance(self.iri, BNode):
            return self.iri

        return URIRef(self.iri)

    def query(
        self,
        query_text: str,
        **kwargs: SPARQLQueryArgument,
    ) -> QueryResult:
        """SPARQL query."""
        return self.ldflex.query(
            query_text=query_text,
            **kwargs,
        )

    def render(
        self,
        node: Union[str, Node],
        as_datatype: NotLiteralNode,
    ) -> Any:
        """Shortcut to render something via iolanta."""
        rendered, stack = self.iolanta.render(
            node=node,
            as_datatype=as_datatype,
        )

        self.stack_children.append(stack)

        return rendered

    def render_all(
        self,
        node: Node,
        as_datatype: NotLiteralNode,
    ) -> Iterable[Any]:
        """Render all we can."""
        return self.iolanta.render_all(node=node, as_datatype=as_datatype)

    def stored_query(self, file_name: str, **kwargs: SPARQLQueryArgument):
        """Execute a stored SPARQL query."""
        query_text = (self.stored_queries_path / file_name).read_text()
        return self.query(
            query_text=query_text,
            **kwargs,
        )

    def show(self) -> FacetOutput:
        """Render the facet."""
        raise NotImplementedError()

    @property
    def language(self) -> Literal:
        """Preferred language for Iolanta output."""
        return self.iolanta.language

    @property
    def stack(self):
        """
        Return stack.

        FIXME: I have no idea. Maybe delete this nonsense.
        """
        return Stack(
            node=self.iri,
            facet=self,
            children=self.stack_children,
        )

    def find_triple(
        self,
        triple: TripleTemplate,
    ) -> Triple | None:
        """Lightweight procedure to find a triple by template."""
        return self.iolanta.find_triple(triple_template=triple)

    @cached_property
    def logger(self):
        """Logger."""
        return self.iolanta.logger.bind(facet=self.__class__.__name__)

    def __str__(self):
        """Render."""
        return str(self.show())
